from microbit import i2c, pin1, pin2, pin8, pin12, pin13,pin14, pin15, sleep
import machine
import music

_speedPercent = 50
_powerByteL = 40
_powerByteR = 40
_motorState = bytearray(5)
_servoBytes = bytearray(2)
_powerBytesLUT = bytes(b'\x00\x0b\x0b\x0c\x0c\x0d\x0d\x0d\x0e\x0e\x0f\x0f\x0f\x10\x10\x11\x11\x11\x12\x12\x13\x13\x13\x14\x14\x15\x15\x15\x16\x16\x17\x17\x17\x18\x19\x1a\x1b\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x36\x38\x3a\x3c\x3f\x41\x44\x46\x49\x4c\x4f\x53\x56\x5a\x5e\x62\x67\x6b\x70\x75\x7b\x81\x87\x8d\x94\x9b\xa3\xab\xb4\xbd\xc6\xd0\xdb\xe6\xf2\xff')
_powerOffset = 0
_powerDifferential = 0
_arcScaling = 0

_alarmSequence = ['c5:1','r','c5:1','r:3']
_UNCONNECTEDERRORMSG = "Please switch robot on."

def _setMotors(dirL,powerL,dirR,powerR):
    global _motorState
    _motorState[1] = dirL
    _motorState[2] = powerL
    _motorState[3] = dirR
    _motorState[4] = powerR
    try:
        i2c.write(0x10,_motorState)
    except:
        raise RuntimeError(_UNCONNECTEDERRORMSG)

def _setSingleMotor(side,dir,power):
    global _motorState
    _motorState[1 + side]= dir
    _motorState[2 + side]= power
    try:
        i2c.write(0x10,_motorState)
    except:
        raise RuntimeError(_UNCONNECTEDERRORMSG)

def _getPowerByte(speed,offset):
    if speed <= 0:
        return 0
    if speed <= 33:
        return int(0.4 * speed + 11 + offset)
    elif speed <= 63:
        return int(0.89 * speed - 5 + offset)
    elif speed < 100:
        return int(min(1.0561 ** speed + 20 + offset, 255))
    else:
        return 255

def _getPowerByteLUT(speed,offset):
    return min(_powerBytesLUT[speed]+ offset,255)

def _getArcBytes(r):
    outerSpeed = _speedPercent
    rCm = int(r * 100)
    threshold = outerSpeed - max(rCm + 20,40)
    if threshold <= 0:
        outerSpeed = min(max(rCm + 40, 40),100)
    reducedSpeed = 0
    if rCm >= 4:
        flattening = (100- outerSpeed)// 2
        reducedSpeed = (rCm * 10 - 35) / (rCm * (11+ (_arcScaling-4)/10)+ 90 + flattening)
        reducedSpeed = reducedSpeed * outerSpeed
    innerByte = _getPowerByteLUT(int(reducedSpeed),0)
    outerByte = _getPowerByteLUT(int(outerSpeed),0)
    return (innerByte,outerByte)

def calibrate(offset, differential=0,arcScaling=0):
    global _powerDifferential
    global _powerOffset
    global _arcScaling
    _powerOffset = max(min(int(offset),50), -10)
    _powerDifferential = max(min(int(differential),150), -150)
    _arcScaling = max(min(arcScaling,50), -15)
    setSpeed(_speedPercent)

def setSpeed(speed):
    global _speedPercent
    global _powerByteL
    global _powerByteR
    _speedPercent = int(min(max(speed,0), 100))
    powerByte = _getPowerByteLUT(_speedPercent,_powerOffset)
    boost = round((1 - _speedPercent / 100) * abs(_powerDifferential)) if _speedPercent > 0 else 0
    reduction = round((_speedPercent/ 100)* abs(_powerDifferential))
    if _powerDifferential > 0:
        _powerByteL = powerByte - reduction
        _powerByteR = powerByte + boost
    else:
        _powerByteL = powerByte + boost
        _powerByteR = powerByte - reduction

def stop():
    _setMotors(0,0, 0, 0)

def forward():
    _setMotors(0,_powerByteL,0,_powerByteR)

def backward():
    _setMotors(1,_powerByteL,1,_powerByteR)

def left():
    _setMotors(1,_powerByteL,0, _powerByteR)

def right():
    _setMotors(0,_powerByteL,1,_powerByteR)  

def rightArc(radius):
    inner,outer = _getArcBytes(radius)
    _setMotors(0,outer, 0, inner)

def leftArc(radius):
    inner,outer = _getArcBytes(radius)
    _setMotors(0,inner,0, outer)

class Motor:
    def __init__(self,side):
        self._side = side

    def rotate(self,speed):       
        speedClamped = int(min(max(abs(speed),0), 100))
        power = _getPowerByteLUT(speedClamped,_powerOffset)
        direction = 0 if speed > 0 else 1
        _setSingleMotor(self._side,direction,power)

def setServo(servo,angle):
    global _servoBytes
    if servo == 'S1'or servo == 'P0':
        _servoBytes[0] = 0x14
    elif servo == 'S2'or servo == 'P1':
        _servoBytes[0]= 0x15
    else:
        raise ValueError("Unknown Servo. Please use 'S1' or 'S2'.")
    if angle < 0 or angle > 180:
        raise ValueError("Invalid angle. Must be between 0 and 180")
    _servoBytes[1] = angle
    try:
        i2c.write(0x10,_servoBytes)    
    except:
        raise RuntimeError(_UNCONNECTEDERRORMSG)

class IR():
    def __init__(self,pin): 
        self.pin = pin

    def read_digital(self):
        return self.pin.read_digital()

    def read_analog(self):
        raise NameError("Maqueen Lite does not support reading analog sensor values.")

def getDistance():
    pin1.write_digital(1)
    pin1.write_digital(0)
    p = machine.time_pulse_us(pin2,1, 50000)
    cm = (p >> 6) + (p >> 10)+ (p >> 11) + (p >> 12) + 1
    return max(min(cm,500),2) if cm > 0 else -1

def setLED(state,stateR=None): 
    stateR = stateR if stateR != None else state
    pin8.write_digital(state)
    pin12.write_digital(stateR) 

def setLEDLeft(state):
    pin8.write_digital(state)

def setLEDRight(state):
    pin12.write_digital(state)

def setAlarm(state):
    if state:
        music.play(_alarmSequence,wait=False,loop=True)
    else:
        music.stop()

def beep():
    music.pitch(440,200, wait=False)  

pin2.set_pull(pin2.NO_PULL)
delay = sleep

irLeft = IR(pin13)
irRight = IR(pin14)
motL = Motor(0)
motR = Motor(2)
